#include "fbinit.h"
#include "devctxt.h"
#include "dc_screen.h"

inline static PUINT16 loc_from_coords(PSCREENPRIVDATA priv, INT32 x, INT32 y)
{
  return priv->pdata + (y * priv->pix_per_row) + x;
}

inline static UINT16 native_from_COLORREF(COLORREF cr)
{
  return (UINT16)(((cr << 8) & 0xF800) | ((cr >> 5) & 0x7E0) | ((cr >> 19) & 0x1F));
}

inline static COLORREF COLORREF_from_native(UINT16 cr)
{
  UINT32 tmp = cr;
  return (COLORREF)(((tmp << 19) & 0xF80000) | ((tmp << 5) & 0xFC00) | ((tmp >> 8) & 0xF800));
}

static inline UINT16 apply_rop2(INT32 op, UINT16 disp, UINT16 pen)
{
  switch (op)
  {
    case R2_BLACK:
	  return 0;
	case R2_NOTMERGEPEN:
	  return ~(disp | pen);
	case R2_MASKNOTPEN:
	  return disp & (~pen);
	case R2_NOTCOPYPEN:
	  return ~pen;
	case R2_MASKPENNOT;
	  return (~disp) & pen;
	case R2_NOT:
	  return ~disp;
    case R2_XORPEN:
	  return disp ^ pen;
	case R2_NOTMASKPEN:
	  return ~(disp & pen);
	case R2_MASKPEN:
	  return disp & pen;
	case R2_NOTXORPEN:
	  return ~(disp ^ pen);
    case R2_NOP:
	  return disp;
	case R2_MERGENOTPEN:
	  return disp | (~pen);
    case R2_COPYPEN:
	  return pen;
	case R2_MERGEPENNOT:
	  return (~disp) | pen;
	case R2_MERGEPEN:
	  return disp | pen;
	case R2_WHITE:
	  return (UINT16)(-1);
  }
}

static COLORREF screen_set_pixel(PVOID privdata, INT32 x, INT32 y, COLORREF color, INT32 op)
{
  PSCREENPRIVDATA priv = (PSCREENPRIVDATA)privdata;
  UINT16 pencolor = native_from_COLORREF(color);
  PUINT16 loc = loc_from_coords(priv, x, y);
  UINT16 screen = *loc;
  *loc = apply_rop2(op, screen, pencolor);
  return COLORREF_from_native(screen);
}

static BOOL screen_line(PVOID privdata, INT32 x1, INT32 y1, INT32 x2, INT32 y2, COLORREF color, INT32 op)
{
  PSCREENPRIVDATA priv = (PSCREENPRIVDATA)privdata;
  UINT16 pencolor = native_from_COLORREF(color);
  INT32 dx = x2 - x1;
  INT32 dy = y2 - y1;
  INT32 tmp;
  PUINT16 loc;
  
  if (ABS(dx) < ABS(dy))
  {
    if (y1 > y2)
	{
	  tmp = x1;
	  x1 = x2;
	  x2 = tmp;
	  tmp = y1;
	  y1 = y2;
	  y2 = tmp;
	  dx = -dx;
	  dy = -dy;
	}
	loc = loc_from_coords(priv, x1, y1);
	tmp = x1;
	x1 << 16;
	dx = (dx << 16) / dy;
	while (y1 <= y2)
	{
	  *loc = apply_rop2(op, *loc, pencolor);
	  x1 += dx;
	  ++y1;
	  loc += priv->pix_per_row;
	  if (tmp != (x1 >> 16))
	  {
	    loc += ((x1 >> 16) - tmp);
		tmp = x1 >> 16;
	  }
	}
  }
  else
  {
    if (x1 > x2)
	{
	  tmp = x1;
	  x1 = x2;
	  x2 = tmp;
	  tmp = y1;
	  y1 = y2;
	  y2 = tmp;
	  dx = -dx;
	  dy = -dy;
	}
	loc = loc_from_coords(priv, x1, y1);
	tmp = y1;
    y1 <<= 16;
    dy = dx ? (dy << 16) / dx : 0;
	while (x1 <= x2)
	{
	  *loc = apply_rop2(op, *loc, pencolor);
	  y1 += dy;
	  ++x1;
	  ++loc;
	  if (tmp != (y1 >> 16))
	  {
	    loc += ((((y1 >> 16) - tmp) * priv->pix_per_row);
	    tmp = y1 >> 16;
	  }
	}
  }
}

static const DCFUNTABLE screen_funtable = {
  screen_set_pixel,
  screen_line
};

static void screen_context_destroy(PVOID obj)
{
  PDCTXT pdctxt = (PDCTXT)obj;
  _DC_FinalizeCommon(pdctxt);
  free(pdctxt->privdata);
}

PDCTXT DC_CreateScreenContext(void)
{
  PDCTXT rc;
  PSCREENPRIVDATA priv;
  
  priv = (PSCREENPRIVDATA)malloc(sizeof(SCREENPRIVDATA));
  if (!priv)
    return NULL;
  priv->pix_per_row = Fb_Info->width;
  priv->pdata = Pb_Ptr;
  
  rc = _DC_Allocate(&screen_funtable, priv);
  if (rc)
  {
    rc->hdr.dtor = screen_context_destroy;
	rc->cliprect.left = rc->cliprect.top = 0;
	rc->cliprect.right = Fb_Info->width - 1;
	rc->cliprect.bottom = Fb_Info->height - 1;
  }
  else
    free(priv);
  return rc;
}
